﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Text;
using System.Threading.Tasks;
using System.IO;
using Newtonsoft.Json;
using System.IO.Compression;

namespace MontoyaTech.Rest.Net
{
    /// <summary>
    /// A set of extensions to help with HttpListenerRequests.
    /// </summary>
    public static class HttpListenerRequestExtensions
    {
        /// <summary>
        /// Reads the content of a HttpListenerRequest as a string and returns it.
        /// </summary>
        /// <param name="request"></param>
        /// <returns></returns>
        public static string ReadAsString(this HttpListenerRequest request)
        {
            try
            {
                //If the request has been compressed, automatically handle it.
                if (request.Headers["Content-Encoding"] == "gzip")
                {
                    using (var inputStream = request.InputStream)
                    using (var gzipStream = new GZipStream(inputStream, CompressionMode.Decompress, true))
                    using (var stream = new StreamReader(gzipStream))
                        return stream.ReadToEnd();
                }
                else
                {
                    using (var input = request.InputStream)
                    using (var stream = new StreamReader(input))
                        return stream.ReadToEnd();
                }
            }
            catch
            {
                return "";
            }
        }

        /// <summary>
        /// Reads the content of a HttpListenerRequest as a json object and returns it.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="request"></param>
        /// <returns></returns>
        public static T ReadAsJson<T>(this HttpListenerRequest request)
        {
            try
            {
                //If the request has been compressed, automatically handle it.
                if (request.Headers["Content-Encoding"] == "gzip")
                {
                    using (var inputStream = request.InputStream)
                    using (var gzipStream = new GZipStream(inputStream, CompressionMode.Decompress, true))
                    using (var stream = new StreamReader(gzipStream))
                        return JsonConvert.DeserializeObject<T>(stream.ReadToEnd());
                }
                else
                {
                    using (var inputStream = request.InputStream)
                    using (var stream = new StreamReader(inputStream))
                        return JsonConvert.DeserializeObject<T>(stream.ReadToEnd());
                }
            }
            catch
            {
                return default(T);
            }
        }

        /// <summary>
        /// Reads the content of a HttpListenerRequest as a byte array.
        /// </summary>
        /// <param name="request"></param>
        /// <returns></returns>
        public static byte[] ReadAsBytes(this HttpListenerRequest request)
        {
            try
            {
                //If the request has been compressed, automatically handle it.
                if (request.Headers["Content-Encoding"] == "gzip")
                {
                    using (var inputStream = request.InputStream)
                    {
                        using (var gzipStream = new GZipStream(inputStream, CompressionMode.Decompress, true))
                        {
                            using (var memoryStream = new MemoryStream())
                            {
                                gzipStream.CopyTo(memoryStream);

                                return memoryStream.ToArray();
                            }
                        }
                    }
                }
                else
                {
                    using (var input = request.InputStream)
                    {
                        using (var stream = new MemoryStream())
                        {
                            input.CopyTo(stream);
                            return stream.ToArray();
                        }
                    }
                }
            }
            catch
            {
                return null;
            }
        }

        /// <summary>
        /// Reads the content of a HttpListenerRequest to the given stream and returns whether or not it succeeded.
        /// </summary>
        /// <param name="request"></param>
        /// <param name="stream"></param>
        /// <returns>Whether or not the read was successful.</returns>
        public static bool ReadToStream(this HttpListenerRequest request, Stream stream)
        {
            try
            {
                //If the request has been compressed, automatically handle it.
                if (request.Headers["Content-Encoding"] == "gzip")
                {
                    using (var inputStream = request.InputStream)
                    {
                        using (var gzipStream = new GZipStream(inputStream, CompressionMode.Decompress, true))
                        {
                            gzipStream.CopyTo(stream);
                        }
                    }
                }
                else
                {
                    using (var input = request.InputStream)
                        input.CopyTo(stream);
                }

                return true;
            }
            catch
            {
                return false;
            }
        }

        /// <summary>
        /// Reads the content of a HttpListenerRequest as http form data and returns it as a dictionary. Any duplicate keys are ignored.
        /// </summary>
        /// <param name="request"></param>
        /// <returns>A dictionary of keyvalue pairs from the form data.</returns>
        public static Dictionary<string, string> ReadAsForm(this HttpListenerRequest request)
        {
            try
            {
                using (var input = request.InputStream)
                using (var stream = new StreamReader(input))
                    return stream.ReadToEnd().Trim().Split('&').Where(value => value.Contains('=')).ToLookup(value => value.Split('=')[0].Trim(), value => value.Split('=')[1].Trim()).ToDictionary(group => group.Key, group => group.FirstOrDefault());
            }
            catch
            {
                return new Dictionary<string, string>();
            }
        }
    }
}
