﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;
using System.IO;
using System.Threading.Tasks;
using System.Xml.Linq;
using System.IO.Pipes;

namespace MontoyaTech.Rest.Net
{
    /// <summary>
    /// The outline of a Rest Client Generator that can generate a C# Client.
    /// </summary>
    public class RestCSharpClientGenerator : RestClientGenerator
    {
        /// <summary>
        /// Whether or not to generate static code, if true the client will be static.
        /// </summary>
        public bool StaticCode = false;

        /// <summary>
        /// Generates a CSharp Client from a given set of routes and returns it.
        /// </summary>
        /// <param name="routes"></param>
        /// <returns></returns>
        public override string Generate(List<Route> routes)
        {
            var includedTypes = this.FindRoutesDependencies(routes);

            var routeGroups = this.FindRouteGroups(routes);

            var writer = new CodeWriter();

            writer.WriteLine($"//Generated using MontoyaTech.Rest.Net - {DateTime.Now.ToShortDateString()}");

            writer.WriteBreak().WriteLine($"public class {this.ClientName}").WriteLine("{").Indent();

            //Create the base url field
            if (this.StaticCode)
                writer.WriteBreak().WriteLine("public static string BaseUrl;");
            else
                writer.WriteBreak().WriteLine("public string BaseUrl;");

            //Create the cookie container field
            if (this.StaticCode)
                writer.WriteBreak().WriteLine("public static System.Net.CookieContainer CookieContainer;");
            else
                writer.WriteBreak().WriteLine("public System.Net.CookieContainer CookieContainer;");

            //Create the client handler field
            if (this.StaticCode)
                writer.WriteBreak().WriteLine("public static System.Net.Http.HttpMessageHandler MessageHandler;");
            else
                writer.WriteBreak().WriteLine("public System.Net.Http.HttpMessageHandler MessageHandler;");

            //Create the http client field
            if (this.StaticCode)
                writer.WriteBreak().WriteLine("public static System.Net.Http.HttpClient HttpClient;");
            else
                writer.WriteBreak().WriteLine("public System.Net.Http.HttpClient HttpClient;");

            //Create the request handler field
            if (this.StaticCode)
                writer.WriteBreak().WriteLine("public static System.Action<System.Net.Http.HttpRequestMessage> RequestHandler;");
            else
                writer.WriteBreak().WriteLine("public System.Action<System.Net.Http.HttpRequestMessage> RequestHandler;");

            //Create fields foreach route group so they can be accessed.
            if (!this.StaticCode)
                foreach (var group in routeGroups)
                    writer.WriteBreak().WriteLine($"public {group.Key}Api {group.Key};");

            //Create the client constructor or init method
            if (this.StaticCode)
            {
                writer.WriteBreak().WriteLine("public static void Init(string baseUrl, System.Net.Http.HttpMessageHandler messageHandler = null, System.Action<System.Net.Http.HttpRequestMessage> requestHandler = null)").WriteLine("{").Indent();

                //Make sure the base url isn't null or whitespace
                writer.WriteBreak().WriteLine("if (string.IsNullOrWhiteSpace(baseUrl))");
                writer.Indent().WriteLine(@"throw new System.ArgumentException(""baseUrl must not be null or whitespace."");").Outdent();

                //If the baseUrl ends with a /, remove it.
                writer.WriteBreak().WriteLine("if (baseUrl.EndsWith('/'))");
                writer.Indent().WriteLine("baseUrl = baseUrl.Substring(0, baseUrl.Length - 1);").Outdent();

                //Init the base url
                writer.WriteBreak().WriteLine($"{this.ClientName}.BaseUrl = baseUrl;");

                //Init the cookie container
                writer.WriteBreak().WriteLine($"{this.ClientName}.CookieContainer = new System.Net.CookieContainer();");

                //Init the client handler
                writer.WriteBreak().WriteLine("if (messageHandler == null)");
                writer.WriteLine("{").Indent();
                writer.WriteLine($"messageHandler = new System.Net.Http.HttpClientHandler()");
                writer.WriteLine("{").Indent();
                writer.WriteLine("AllowAutoRedirect = true,");
                writer.WriteLine("UseCookies = true,");
                writer.WriteLine($"CookieContainer = {this.ClientName}.CookieContainer,");
                writer.WriteLine("AutomaticDecompression = System.Net.DecompressionMethods.GZip | System.Net.DecompressionMethods.Deflate");
                writer.Outdent().WriteLine("};");
                writer.Outdent().WriteLine("}");

                //Store the message handler
                writer.WriteBreak().WriteLine($"{this.ClientName}.MessageHandler = messageHandler;");

                //Store the request handler
                writer.WriteBreak().WriteLine($"{this.ClientName}.RequestHandler = requestHandler;");

                //Init the http client
                writer.WriteBreak().WriteLine($"{this.ClientName}.HttpClient = new System.Net.Http.HttpClient(handler);");
                writer.WriteBreak().WriteLine(@$"{this.ClientName}.HttpClient.DefaultRequestHeaders.Add(""Accept"", ""*/*"");");
                writer.WriteBreak().WriteLine(@$"{this.ClientName}.HttpClient.DefaultRequestHeaders.Add(""Connection"", ""keep-alive"");");
                writer.WriteBreak().WriteLine(@$"{this.ClientName}.HttpClient.DefaultRequestHeaders.Add(""Accept-Encoding"", ""identity"");");

                writer.Outdent().WriteLine("}");
            }
            else
            {
                writer.WriteBreak().WriteLine($"public {this.ClientName}(string baseUrl, System.Net.Http.HttpMessageHandler messageHandler = null, System.Action<System.Net.Http.HttpRequestMessage> requestHandler = null)").WriteLine("{").Indent();

                //Make sure the base url isn't null or whitespace
                writer.WriteBreak().WriteLine("if (string.IsNullOrWhiteSpace(baseUrl))");
                writer.Indent().WriteLine(@"throw new System.ArgumentException(""baseUrl must not be null or whitespace."");").Outdent();

                //If the baseUrl ends with a /, remove it.
                writer.WriteBreak().WriteLine("if (baseUrl.EndsWith('/'))");
                writer.Indent().WriteLine("baseUrl = baseUrl.Substring(0, baseUrl.Length - 1);").Outdent();

                //Init the base url
                writer.WriteBreak().WriteLine("this.BaseUrl = baseUrl;");

                //Init the cookie container
                writer.WriteBreak().WriteLine("this.CookieContainer = new System.Net.CookieContainer();");

                //Init the client handler
                writer.WriteBreak().WriteLine("if (messageHandler == null)");
                writer.WriteLine("{").Indent();
                writer.WriteLine("messageHandler = new System.Net.Http.HttpClientHandler()");
                writer.WriteLine("{").Indent();
                writer.WriteLine("AllowAutoRedirect = true,");
                writer.WriteLine("UseCookies = true,");
                writer.WriteLine("CookieContainer = this.CookieContainer,");
                writer.WriteLine("AutomaticDecompression = System.Net.DecompressionMethods.GZip | System.Net.DecompressionMethods.Deflate");
                writer.Outdent().WriteLine("};");
                writer.Outdent().WriteLine("}");

                //Store the message handler
                writer.WriteBreak().WriteLine("this.MessageHandler = messageHandler;");

                //Store the request handler
                writer.WriteBreak().WriteLine("this.RequestHandler = requestHandler;");

                //Init the http client
                writer.WriteBreak().WriteLine("this.HttpClient = new System.Net.Http.HttpClient(handler);");
                writer.WriteBreak().WriteLine(@"this.HttpClient.DefaultRequestHeaders.Add(""Accept"", ""*/*"");");
                writer.WriteBreak().WriteLine(@"this.HttpClient.DefaultRequestHeaders.Add(""Connection"", ""keep-alive"");");
                writer.WriteBreak().WriteLine(@"this.HttpClient.DefaultRequestHeaders.Add(""Accept-Encoding"", ""identity"");");

                //Init all the route group fields
                foreach (var group in routeGroups)
                    writer.WriteBreak().WriteLine($"this.{group.Key} = new {group.Key}Api(this);");

                writer.Outdent().WriteLine("}");
            }

            this.GenerateCSharpRouteGroups(routeGroups, writer);

            this.GenerateCSharpIncludedTypes(includedTypes, writer);

            writer.Outdent().WriteLine("}");

            return writer.ToString();
        }

        /// <summary>
        /// Generates C# for a set of included types.
        /// </summary>
        /// <param name="types"></param>
        /// <param name="writer"></param>
        protected internal virtual void GenerateCSharpIncludedTypes(List<Type> types, CodeWriter writer)
        {
            foreach (var type in types)
            {
                bool subType = false;

                //See if this type belongs to another type in the list.
                if (type.DeclaringType != null)
                    for (int i = 0; i < types.Count && !subType; i++)
                        if (type.DeclaringType == types[i])
                            subType = true;

                //If not, generate the C# for this type.
                if (!subType)
                    this.GenerateCSharpIncludedType(type, types, writer);
            }
        }

        /// <summary>
        /// Generates C# for a given included type.
        /// </summary>
        /// <param name="type"></param>
        /// <param name="writer"></param>
        protected internal virtual void GenerateCSharpIncludedType(Type type, List<Type> types, CodeWriter writer)
        {
            writer.WriteBreak();

            var newName = type.GetCustomAttribute<RouteTypeName>();

            writer.Write($"public {(type.IsEnum ? "enum" : "class")} {(newName != null ? newName.Name : type.Name)}");

            if (type.IsEnum)
                writer.Write(" : ").Write(this.GetTypeFullyResolvedName(Enum.GetUnderlyingType(type)));
            else if (!(IsTypeDotNet(type.BaseType) && type.BaseType.Name == "Object"))
                writer.Write(" : ").Write(this.GetTypeFullyResolvedName(type.BaseType));

            writer.NewLine().WriteLine("{").Indent();

            FieldInfo[] fields;
            
            if (type.IsEnum)
                fields = type.GetFields(BindingFlags.DeclaredOnly | BindingFlags.Public | BindingFlags.Static);
            else
                fields = type.GetFields(BindingFlags.DeclaredOnly | BindingFlags.Instance | BindingFlags.Public);

            if (fields != null)
                foreach (var field in fields)
                    if (field.IsPublic && !field.IsSpecialName)
                        this.GenerateCSharpIncludedField(field, writer);

            var properties = type.GetProperties(BindingFlags.DeclaredOnly | BindingFlags.Instance | BindingFlags.Public);

            if (properties != null)
                foreach (var property in properties)
                    if (!property.IsSpecialName && property.GetSetMethod() != null && property.GetGetMethod() != null)
                        this.GenerateCSharpIncludedProperty(property, writer);

            if (!type.IsEnum)
            {
                //Generate an empty constructor
                writer.WriteBreak();
                writer.WriteLine($"public {(newName != null ? newName.Name : type.Name)}() {{ }}");

                //Generate a constructor to set all the fields/properties from an existing instance
                writer.WriteBreak();
                writer.WriteLine($"public {(newName != null ? newName.Name : type.Name)}({(newName != null ? newName.Name : type.Name)} instance)");
                writer.WriteLine('{').Indent();

                foreach (var field in fields)
                    writer.WriteLine($"this.{field.Name} = instance.{field.Name};");

                foreach (var property in properties)
                    writer.WriteLine($"this.{property.Name} = instance.{property.Name};");

                writer.Outdent().WriteLine('}');

                //Generate a constructor to set all the fields/properties with optional default values
                if (fields.Length > 0 || properties.Length > 0)
                {
                    writer.WriteBreak();
                    writer.Write($"public {(newName != null ? newName.Name : type.Name)}(");

                    foreach (var field in fields)
                        writer.WriteSeparator().Write($"{this.GetTypeFullyResolvedName(field.FieldType)} {field.Name} = {this.GetTypeDefaultValue(field.FieldType)}");

                    foreach (var property in properties)
                        writer.WriteSeparator().Write($"{this.GetTypeFullyResolvedName(property.PropertyType)} {property.Name} = {this.GetTypeDefaultValue(property.PropertyType)}");

                    writer.WriteLine(")");
                    writer.WriteLine('{').Indent();

                    foreach (var field in fields)
                        writer.WriteLine($"this.{field.Name} = {field.Name};");

                    foreach (var property in properties)
                        writer.WriteLine($"this.{property.Name} = {property.Name};");

                    writer.Outdent().WriteLine('}');
                }
            }

            //Generate C# for any types that belong to this one.
            for (int i = 0; i < types.Count; i++)
                if (types[i].DeclaringType == type)
                    GenerateCSharpIncludedType(types[i], types, writer);

            writer.Outdent().WriteLine("}");
        }

        /// <summary>
        /// Generates C# for a field inside an included type.
        /// </summary>
        /// <param name="field"></param>
        /// <param name="writer"></param>
        protected internal virtual void GenerateCSharpIncludedField(FieldInfo field, CodeWriter writer)
        {
            writer.WriteBreak();

            if (field.DeclaringType != null && field.DeclaringType.IsEnum)
                writer.WriteLine($"{field.Name} = {field.GetRawConstantValue()},");
            else
                writer.WriteLine($"public {this.GetTypeFullyResolvedName(field.FieldType)} {field.Name};");
        }

        /// <summary>
        /// Generates C# for a property inside an included type.
        /// </summary>
        /// <param name="property"></param>
        /// <param name="writer"></param>
        protected internal virtual void GenerateCSharpIncludedProperty(PropertyInfo property, CodeWriter writer)
        {
            writer.WriteBreak();

            writer.WriteLine($"public {this.GetTypeFullyResolvedName(property.PropertyType)} {property.Name} {{ get; set; }}");
        }

        /// <summary>
        /// Generates C# for a set of route groups.
        /// </summary>
        /// <param name="groups"></param>
        /// <param name="writer"></param>
        protected internal virtual void GenerateCSharpRouteGroups(Dictionary<string, List<Route>> groups, CodeWriter writer)
        {
            foreach (var group in groups)
                this.GenerateCSharpRouteGroup(group.Key, group.Value, writer);
        }

        /// <summary>
        /// Generates C# for a given route group.
        /// </summary>
        /// <param name="name"></param>
        /// <param name="routes"></param>
        /// <param name="writer"></param>
        protected internal virtual void GenerateCSharpRouteGroup(string name, List<Route> routes, CodeWriter writer)
        {
            writer.WriteBreak();

            //Output the class header
            if (this.StaticCode)
                writer.WriteLine($"public class {name}").WriteLine("{").Indent();
            else
                writer.WriteLine($"public class {name}Api").WriteLine("{").Indent();

            //Output the client instance
            if (!this.StaticCode)
                writer.WriteBreak().WriteLine($"public {this.ClientName} Client;");


            //Output the constuctor if not static.
            if (!this.StaticCode)
            {
                writer.WriteBreak().WriteLine($"public {name}Api({this.ClientName} client)").WriteLine("{").Indent();

                writer.WriteLine("this.Client = client;");

                writer.Outdent().WriteLine("}");
            }

            //Output all the route functions.
            foreach (var route in routes)
                this.GenerateCSharpRouteFunction(route, writer);

            writer.Outdent().WriteLine("}");
        }

        /// <summary>
        /// Generates a C# function for a given route.
        /// </summary>
        /// <param name="route"></param>
        /// <param name="writer"></param>
        /// <exception cref="NotSupportedException"></exception>
        protected internal virtual void GenerateCSharpRouteFunction(Route route, CodeWriter writer)
        {
            writer.WriteBreak();

            var methodInfo = route.GetTarget().GetMethodInfo();

            var routeName = methodInfo.GetCustomAttribute<RouteName>();

            var routeRequest = methodInfo.GetCustomAttribute<RouteRequest>();

            var routeResponse = methodInfo.GetCustomAttribute<RouteResponse>();

            //Generate the route function header
            if (this.StaticCode)
                writer.Write($"public static {(routeResponse == null ? "void" : (routeResponse.Dynamic ? "dynamic" : this.GetTypeFullyResolvedName(routeResponse.ResponseType)))} {(routeName == null ? methodInfo.Name : routeName.Name)}(");
            else
                writer.Write($"public {(routeResponse == null ? "void" : (routeResponse.Dynamic ? "dynamic" : this.GetTypeFullyResolvedName(routeResponse.ResponseType)))} {(routeName == null ? methodInfo.Name : routeName.Name)}(");

            //Generate the functions parameters
            var parameters = methodInfo.GetParameters();

            if (parameters != null)
            {
                for (int i = 1; i < parameters.Length; i++)
                {
                    writer.WriteSeparator();
                    writer.Write(this.GetTypeFullyResolvedName(parameters[i].ParameterType)).Write(" ").Write(parameters[i].Name);
                }
            }

            if (routeRequest != null)
            {
                writer.WriteSeparator();

                if (routeRequest.Dynamic)
                    writer.Write("dynamic");
                else
                    writer.Write(this.GetTypeFullyResolvedName(routeRequest.RequestType));

                writer.Write(" request");
            }

            if (routeResponse != null && routeResponse.Parameter)
            {
                writer.WriteSeparator();

                if (routeResponse.Dynamic)
                    writer.Write("dynamic");
                else
                    writer.Write(this.GetTypeFullyResolvedName(routeResponse.ResponseType));

                writer.Write(" input");
            }

            //If JSON, add a compress parameter to control compressing the content.
            if (routeRequest != null && routeRequest.Json)
            {
                writer.WriteSeparator();
                writer.Write("bool compress = false");
            }

            writer.WriteLine(")").WriteLine("{").Indent();

            //Generate the message code
            writer.WriteBreak().Write($"var message = new System.Net.Http.HttpRequestMessage(");

            switch (route.Method.ToLower())
            {
                case "post":
                    writer.Write("System.Net.Http.HttpMethod.Post");
                    break;
                case "get":
                    writer.Write("System.Net.Http.HttpMethod.Get");
                    break;
                case "delete":
                    writer.Write("System.Net.Http.HttpMethod.Delete");
                    break;
                case "put":
                    writer.Write("System.Net.Http.HttpMethod.Put");
                    break;
                case "options":
                    writer.Write("System.Net.Http.HttpMethod.Options");
                    break;
                case "patch":
                    writer.Write("System.Net.Http.HttpMethod.Patch");
                    break;
                case "head":
                    writer.Write("System.Net.Http.HttpMethod.Head");
                    break;
                case "trace":
                    writer.Write("System.Net.Http.HttpMethod.Trace");
                    break;
                default:
                    throw new NotSupportedException("Unsupport route method:" + route.Method);
            }

            if (this.StaticCode)
                writer.WriteSeparator().Write('$').Write('"').Write($"{{{this.ClientName}.BaseUrl}}");
            else
                writer.WriteSeparator().Write('$').Write('"').Write("{this.Client.BaseUrl}");

            //Reconstruct the route syntax into a request url.
            var components = route.Syntax.Split('/');
            int argumentIndex = 0;
            foreach (var component in components)
            {
                if (writer.Peek() != '/')
                    writer.Write('/');

                if (!string.IsNullOrWhiteSpace(component))
                {
                    if (component.StartsWith("{"))
                    {
                        writer.Write("{").Write(parameters[argumentIndex++ + 1].Name).Write("}");
                    }
                    else if (component == "*")
                    {
                        writer.Write("*");
                    }
                    else if (component == "**")
                    {
                        break;
                    }
                    else
                    {
                        writer.Write(component);
                    }
                }
            }

            writer.Write('"').WriteLine(");");

            //Invoke the request handler if needed.
            if (this.StaticCode)
                writer.WriteBreak().WriteLine($"{this.ClientName}.RequestHandler?.Invoke(message);");
            else
                writer.WriteBreak().WriteLine($"this.Client.RequestHandler?.Invoke(message);");

            //Add the request content if any.
            if (routeRequest != null)
            {
                if (routeRequest.RequestType.IsAssignableTo(typeof(Stream)))
                {
                    writer.WriteBreak().WriteLine("request.Seek(0, System.IO.SeekOrigin.Begin);");

                    writer.WriteBreak().WriteLine("message.Content = new System.Net.Http.StreamContent(request);");
                }
                else if (routeRequest.Json)
                {
                    writer.WriteBreak().WriteLine("if (compress)").WriteLine("{").Indent();

                    writer.WriteLine("using (var uncompressedStream = new System.IO.MemoryStream(System.Text.Encoding.UTF8.GetBytes(Newtonsoft.Json.JsonConvert.SerializeObject(request))))");
                    writer.WriteLine("{").Indent();
                    writer.WriteLine("using (var compressedStream = new System.IO.MemoryStream())");
                    writer.WriteLine("{").Indent();
                    writer.WriteLine("using (var gzipStream = new System.IO.Compression.GZipStream(compressedStream, System.IO.Compression.CompressionMode.Compress, true))");
                    writer.Indent().WriteLine("uncompressedStream.CopyTo(gzipStream);").Outdent();

                    writer.WriteBreak();
                    writer.WriteLine("message.Content = new System.Net.Http.ByteArrayContent(compressedStream.ToArray());");
                    writer.WriteLine("message.Content.Headers.ContentType = new System.Net.Http.Headers.MediaTypeHeaderValue(System.Net.Mime.MediaTypeNames.Application.Json);");
                    writer.WriteLine("message.Content.Headers.ContentEncoding.Add(\"gzip\");");

                    writer.Outdent().WriteLine("}");
                    writer.Outdent().WriteLine("}");

                    writer.Outdent().WriteLine("}");
                    writer.WriteLine("else").WriteLine("{").Indent();

                    writer.WriteBreak().WriteLine("message.Content = new System.Net.Http.StringContent(Newtonsoft.Json.JsonConvert.SerializeObject(request));");
                    writer.WriteBreak().WriteLine("message.Content.Headers.ContentType = new System.Net.Http.Headers.MediaTypeHeaderValue(System.Net.Mime.MediaTypeNames.Application.Json);");

                    writer.Outdent().WriteLine("}");
                }
                else
                {
                    writer.WriteBreak().WriteLine("message.Content = new System.Net.Http.StringContent(request.ToString());");
                }
            }

            //Generate the response code
            if (this.StaticCode)
                writer.WriteBreak().WriteLine($"var response = {this.ClientName}.HttpClient.Send(message, System.Net.Http.HttpCompletionOption.ResponseHeadersRead);");
            else
                writer.WriteBreak().WriteLine("var response = this.Client.HttpClient.Send(message, System.Net.Http.HttpCompletionOption.ResponseHeadersRead);");

            //Handle the response
            if (routeResponse != null)
            {
                writer.WriteBreak().WriteLine("if (response.IsSuccessStatusCode)").WriteLine("{").Indent();

                if (routeResponse.ResponseType != null && routeResponse.ResponseType.IsAssignableTo(typeof(Stream)))
                {
                    if (routeResponse.Parameter)
                    {
                        writer.WriteBreak().WriteLine("response.Content.CopyToAsync(input).GetAwaiter().GetResult();");

                        writer.WriteBreak().WriteLine("return input;");
                    }
                    else
                    {
                        writer.WriteBreak().WriteLine($"var stream = new {this.GetTypeFullyResolvedName(routeResponse.ResponseType)}();");

                        writer.WriteBreak().WriteLine("response.Content.CopyToAsync(stream).GetAwaiter().GetResult();");

                        writer.WriteBreak().WriteLine("return stream;");
                    }
                }
                else
                {
                    writer.WriteBreak().WriteLine("var content = response.Content.ReadAsStringAsync().GetAwaiter().GetResult();");

                    writer.WriteBreak().WriteLine("if (string.IsNullOrEmpty(content))").Indent().WriteLine("return default;").Outdent();

                    if (routeResponse.Json)
                    {
                        writer.WriteBreak().WriteLine($"return Newtonsoft.Json.JsonConvert.DeserializeObject<{(routeResponse.Dynamic ? "dynamic" : this.GetTypeFullyResolvedName(routeResponse.ResponseType))}>(content);");
                    }
                    else
                    {
                        switch (Type.GetTypeCode(routeResponse.ResponseType))
                        {
                            case TypeCode.Boolean:
                                writer.WriteBreak().WriteLine("return bool.Parse(content);");
                                break;

                            case TypeCode.Byte:
                                writer.WriteBreak().WriteLine("return byte.Parse(content);");
                                break;

                            case TypeCode.Char:
                                writer.WriteBreak().WriteLine("return content[0];");
                                break;

                            case TypeCode.DateTime:
                                writer.WriteBreak().WriteLine("return System.DateTime.Parse(content);");
                                break;

                            case TypeCode.Decimal:
                                writer.WriteBreak().WriteLine("return decimal.Parse(content);");
                                break;

                            case TypeCode.Double:
                                writer.WriteBreak().WriteLine("return double.Parse(content);");
                                break;

                            case TypeCode.Int16:
                                writer.WriteBreak().WriteLine("return short.Parse(content);");
                                break;

                            case TypeCode.Int32:
                                writer.WriteBreak().WriteLine("return int.Parse(content);");
                                break;

                            case TypeCode.Int64:
                                writer.WriteBreak().WriteLine("return long.Parse(content);");
                                break;

                            case TypeCode.SByte:
                                writer.WriteBreak().WriteLine("return sbyte.Parse(content);");
                                break;

                            case TypeCode.Single:
                                writer.WriteBreak().WriteLine("return float.Parse(content);");
                                break;

                            case TypeCode.String:
                                writer.WriteBreak().WriteLine("return content;");
                                break;

                            case TypeCode.UInt16:
                                writer.WriteBreak().WriteLine("return ushort.Parse(content);");
                                break;

                            case TypeCode.UInt32:
                                writer.WriteBreak().WriteLine("return uint.Parse(content);");
                                break;

                            case TypeCode.UInt64:
                                writer.WriteBreak().WriteLine("return ulong.Parse(content);");
                                break;

                            case TypeCode.Object:
                                throw new NotSupportedException("RouteResponse has JSON=false but ResponseType is an object.");
                        }
                    }
                }

                writer.Outdent().WriteLine("}").WriteLine("else").WriteLine("{").Indent();
                writer.WriteLine(@"throw new System.Exception(""Unexpected Http Response StatusCode:"" + response.StatusCode);");
                writer.Outdent().WriteLine("}");
            }
            else
            {
                writer.WriteBreak().WriteLine("if (!response.IsSuccessStatusCode)").Indent();
                writer.WriteLine(@"throw new System.Exception(""Unexpected Http Response StatusCode:"" + response.StatusCode);").Outdent();
            }

            //Close off the route function.
            writer.Outdent().WriteLine("}");
        }
    }
}
