using System;
using System.Collections.Generic;
using System.ComponentModel.DataAnnotations;
using System.ComponentModel.Design;
using System.Data;
using System.Linq;
using System.Net.Http;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;
using System.Xml.Linq;

namespace MontoyaTech.Rest.Net
{
    /// <summary>
    /// A class that can take a set of routes and generate code
    /// for a client that can be used to interact with them.
    /// </summary>
    public class RestClientGenerator
    {
        /// <summary>
        /// The name of the client to generate.
        /// </summary>
        public string ClientName = "Client";

        /// <summary>
        /// Generates a Rest Client from a RouteListener and returns the code.
        /// </summary>
        /// <returns></returns>
        public string Generate(RouteListener listener)
        {
            return this.Generate(listener.Routes);
        }

        /// <summary>
        /// Generates a Rest Client from a set of routes and returns the code.
        /// </summary>
        /// <param name="routes"></param>
        /// <returns></returns>
        public virtual string Generate(List<Route> routes)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Returns whether or not a given type belongs to DotNet.
        /// </summary>
        /// <param name="type"></param>
        /// <returns></returns>
        protected internal virtual bool IsTypeDotNet(Type type)
        {
            var assemblyName = type.Assembly.GetName().Name;

            return assemblyName == "System.Private.CoreLib" || assemblyName == "System.Net.Primitives";
        }

        /// <summary>
        /// Finds all the dependencies for a given type and returns them.
        /// </summary>
        /// <param name="type"></param>
        /// <param name="dependencies"></param>
        /// <returns></returns>
        protected internal virtual HashSet<Type> FindTypeDependencies(Type type, HashSet<Type> dependencies = null)
        {
            if (dependencies == null)
                dependencies = new HashSet<Type>();
            else if (dependencies.Contains(type))
                return dependencies;
            else if (!this.IsTypeDotNet(type))
                dependencies.Add(type);

            var arguments = type.GetGenericArguments();

            if (arguments != null)
                foreach (var argument in arguments)
                    if (argument != type)
                        this.FindTypeDependencies(argument, dependencies);

            if (this.IsTypeDotNet(type))
                return dependencies;

            dependencies.Add(type);

            if (type.IsEnum)
                return dependencies;

            this.FindTypeDependencies(type.BaseType, dependencies);

            var fields = type.GetFields();

            if (fields != null)
                foreach (var field in fields)
                    if (field.IsPublic && !field.IsSpecialName && field.FieldType != type)
                        this.FindTypeDependencies(field.FieldType, dependencies);

            var properties = type.GetProperties();

            if (properties != null)
                foreach (var property in properties)
                    if (!property.IsSpecialName && property.GetSetMethod() != null && property.GetGetMethod() != null && property.PropertyType != type)
                        this.FindTypeDependencies(property.PropertyType, dependencies);

            return dependencies;
        }

        /// <summary>
        /// Finds all the types that a given route depends on to function.
        /// </summary>
        /// <param name="route"></param>
        /// <returns></returns>
        protected internal virtual List<Type> FindRouteDependencies(Route route)
        {
            var dependencies = new HashSet<Type>();

            var methodInfo = route.GetTarget().GetMethodInfo();

            //If this route is hidden, return nothing.
            if (methodInfo.GetCustomAttribute<RouteHidden>() != null)
                return dependencies.ToList();

            if (methodInfo != null)
            {
                var parameters = methodInfo.GetParameters();

                if (parameters != null)
                {
                    for (int i = 1; i < parameters.Length; i++)
                    {
                        var types = this.FindTypeDependencies(parameters[i].ParameterType);

                        foreach (var type in types)
                            if (!dependencies.Contains(type))
                                dependencies.Add(type);
                    }
                }

                var routeRequest = methodInfo.GetCustomAttribute<RouteRequest>();

                if (routeRequest != null && routeRequest.RequestType != null)
                {
                    var types = this.FindTypeDependencies(routeRequest.RequestType);

                    foreach (var type in types)
                        if (!dependencies.Contains(type))
                            dependencies.Add(type);
                }

                var routeResponse = methodInfo.GetCustomAttribute<RouteResponse>();

                if (routeResponse != null && routeResponse.ResponseType != null)
                {
                    var types = this.FindTypeDependencies(routeResponse.ResponseType);

                    foreach (var type in types)
                        if (!dependencies.Contains(type))
                            dependencies.Add(type);
                }

                var routeIncludes = methodInfo.GetCustomAttributes<RouteInclude>();

                if (routeIncludes != null)
                {
                    foreach (var include in routeIncludes)
                    {
                        var types = this.FindTypeDependencies(include.Type);

                        foreach (var type in types)
                            if (!dependencies.Contains(type))
                                dependencies.Add(type);
                    }
                }
            }

            return dependencies.ToList();
        }

        /// <summary>
        /// Finds all the types that a given set of routes depend on to function.
        /// </summary>
        /// <param name="routes"></param>
        /// <returns></returns>
        protected internal virtual List<Type> FindRoutesDependencies(List<Route> routes)
        {
            var dependencies = new HashSet<Type>();

            foreach (var route in routes)
            {
                var types = this.FindRouteDependencies(route);

                if (types != null)
                    for (int i = 0; i < types.Count; i++)
                        if (!dependencies.Contains(types[i]))
                            dependencies.Add(types[i]);
            }

            return dependencies.ToList();
        }

        /// <summary>
        /// Returns the fully resolve name for a given type.
        /// </summary>
        /// <param name="type"></param>
        /// <returns></returns>
        protected internal virtual string GetTypeFullyResolvedName(Type type)
        {
            if (this.IsTypeDotNet(type))
            {
                var typeCode = Type.GetTypeCode(type);

                if (typeCode != TypeCode.Object && !type.IsEnum)
                {
                    switch (typeCode)
                    {
                        case TypeCode.Boolean:
                            return "bool";
                        case TypeCode.Byte:
                            return "byte";
                        case TypeCode.Char:
                            return "char";
                        case TypeCode.DateTime:
                            return "System.DateTime";
                        case TypeCode.Decimal:
                            return "decimal";
                        case TypeCode.Double:
                            return "double";
                        case TypeCode.Int16:
                            return "short";
                        case TypeCode.Int32:
                            return "int";
                        case TypeCode.Int64:
                            return "long";
                        case TypeCode.SByte:
                            return "sbyte";
                        case TypeCode.Single:
                            return "float";
                        case TypeCode.String:
                            return "string";
                        case TypeCode.UInt16:
                            return "ushort";
                        case TypeCode.UInt32:
                            return "uint";
                        case TypeCode.UInt64:
                            return "ulong";
                    }
                }

                var builder = new StringBuilder();

                builder.Append(type.Namespace);

                int genericSymbol = type.Name.IndexOf('`');

                if (genericSymbol == -1)
                    builder.Append(".").Append(type.Name);
                else
                    builder.Append(".").Append(type.Name.Substring(0, genericSymbol));

                var genericArguments = type.GetGenericArguments();

                if (genericArguments != null && genericArguments.Length > 0)
                {
                    builder.Append("<");

                    for (int i = 0; i < genericArguments.Length; i++)
                    {
                        if (i > 0)
                            builder.Append(", ");

                        builder.Append(this.GetTypeFullyResolvedName(genericArguments[i]));
                    }

                    builder.Append(">");
                }

                return builder.ToString();
            }
            else
            {
                var builder = new StringBuilder();

                var newName = type.GetCustomAttribute<RouteTypeName>();

                if (newName != null)
                {
                    builder.Append(newName.Name);
                }
                else
                {
                    int genericSymbol = type.Name.IndexOf('`');

                    if (genericSymbol == -1)
                        builder.Append(type.Name);
                    else
                        builder.Append(type.Name.Substring(0, genericSymbol));
                }

                var genericArguments = type.GetGenericArguments();

                if (genericArguments != null && genericArguments.Length > 0)
                {
                    builder.Append("<");

                    for (int i = 0; i < genericArguments.Length; i++)
                    {
                        if (i > 0)
                            builder.Append(", ");

                        builder.Append(this.GetTypeFullyResolvedName(genericArguments[i]));
                    }

                    builder.Append(">");
                }

                return builder.ToString();
            }
        }

        /// <summary>
        /// Returns the default value for a given type.
        /// </summary>
        /// <param name="type"></param>
        /// <returns></returns>
        protected internal virtual string GetTypeDefaultValue(Type type)
        {
            var typeCode = Type.GetTypeCode(type);

            if (type.IsEnum)
                return "0";
            else if (typeCode == TypeCode.String || typeCode == TypeCode.Object)
                return "null";
            else if (typeCode == TypeCode.Char)
                return "'\\0'";
            else if (typeCode == TypeCode.Boolean)
                return "false";
            else
                return Activator.CreateInstance(type).ToString();
        }

        /// <summary>
        /// Finds all the route groupings from a set of routes and returns those groups.
        /// </summary>
        /// <param name="routes"></param>
        /// <returns></returns>
        protected internal virtual Dictionary<string, List<Route>> FindRouteGroups(List<Route> routes)
        {
            var groups = new Dictionary<string, List<Route>>();

            foreach (var route in routes)
            {
                var methodInfo = route.GetTarget().GetMethodInfo();

                //Skip any routes that are hidden.
                if (methodInfo.GetCustomAttribute<RouteHidden>() != null)
                    continue;

                var routeGroup = methodInfo.GetCustomAttribute<RouteGroup>();

                //If there wasn't a route group on the method, see if there is one on the declaring type.
                if (routeGroup == null)
                    routeGroup = methodInfo.DeclaringType.GetCustomAttribute<RouteGroup>();

                //Group this route.
                string group = (routeGroup != null ? routeGroup.Name : methodInfo.DeclaringType.Name);

                if (groups.ContainsKey(group))
                    groups[group].Add(route);
                else
                    groups.Add(group, new List<Route>() { route });
            }

            return groups;
        }

        /// <summary>
        /// Sorts a list of types by the order of which should be declared first. Useful
        /// for languages that require classes to be dependency ordered.
        /// </summary>
        /// <param name="types"></param>
        /// <returns></returns>
        protected internal virtual List<Type> SortTypesByDependencies(List<Type> types)
        {
            var dependencies = new Dictionary<Type, HashSet<Type>>();

            //Find all the inner dependencies of the types
            foreach (var type in types)
                if (!dependencies.ContainsKey(type))
                    dependencies.Add(type, FindTypeDependencies(type).ToHashSet());

            //Now begin sorting these dependencies
            var sorted = new LinkedList<KeyValuePair<Type, HashSet<Type>>>();
            foreach (var dependency in dependencies)
            {
                var best = sorted.Last;
                var curr = sorted.Last;
                while (curr != null)
                {
                    if (dependency.Value.Contains(curr.Value.Key))
                        break;
                    else if (curr.Value.Value.Contains(dependency.Key))
                        best = curr.Previous;

                    curr = curr.Previous;
                }

                if (best == null)
                    sorted.AddFirst(dependency);
                else
                    best.List.AddAfter(best, dependency);
            }

            return sorted.Select(value => value.Key).ToList();
        }
    }
}
