﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.CompilerServices;
using System.Text;
using System.Threading.Tasks;
using System.IO;
using System.Collections.Concurrent;
using System.Security.Principal;

namespace MontoyaTech.Rest.Net
{
    /// <summary>
    /// The outline of a FileCache that can be used with Routes to speed
    /// up response times.
    /// </summary>
    public class RouteFileCache
    {
        /// <summary>
        /// The outline of a CachedFile.
        /// </summary>
        public class CachedFile
        {
            public string FilePath;

            public byte[] Content;

            public int HitCount;

            public DateTime Cached = DateTime.Now;

            public DateTime LastAccess = DateTime.Now;

            public CachedFile(string filePath, byte[] content)
            {
                this.FilePath = filePath;
                this.Content = content;
            }
        }

        /// <summary>
        /// The files that are currently cached.
        /// </summary>
        public Dictionary<string, CachedFile> CachedFiles = new Dictionary<string, CachedFile>();

        /// <summary>
        /// The current number of bytes cached.
        /// </summary>
        public int BytesCached { get; internal set; }

        /// <summary>
        /// The max bytes that can be cached at a time.
        /// </summary>
        public int MaxBytesCached { get; internal set; }

        /// <summary>
        /// Creates a new RouteFileCache with the max bytes that can be cached.
        /// </summary>
        /// <param name="maxBytes"></param>
        public RouteFileCache(int maxBytes)
        {
            this.MaxBytesCached = maxBytes;
        }

        /// <summary>
        /// Returns whether or not a file was cached and the content of the file if it was.
        /// </summary>
        /// <param name="filePath"></param>
        /// <param name="content"></param>
        /// <returns></returns>
        public bool Cached(string filePath, out byte[] content)
        {
            lock (this.CachedFiles)
            {
                content = null;

                if (!this.CachedFiles.ContainsKey(filePath))
                    return false;

                var cached = this.CachedFiles[filePath];
                cached.HitCount++;
                cached.LastAccess = DateTime.Now;

                content = cached.Content;

                return true;
            }
        }

        /// <summary>
        /// Returns whether or not a file was cached.
        /// </summary>
        /// <param name="filePath"></param>
        /// <returns></returns>
        public bool Cached(string filePath)
        {
            lock (this.CachedFiles)
                return this.CachedFiles.ContainsKey(filePath);
        }

        /// <summary>
        /// Caches a new file and returns the content, if the file was already cached the content will be returned.
        /// </summary>
        /// <param name="filePath"></param>
        /// <returns></returns>
        /// <exception cref="FileNotFoundException"></exception>
        public byte[] Cache(string filePath)
        {
            lock (this.CachedFiles)
            {
                if (!this.CachedFiles.ContainsKey(filePath))
                {
                    if (File.Exists(filePath))
                    {
                        var content = File.ReadAllBytes(filePath);

                        this.Cache(filePath, content);

                        return content;
                    }
                    else
                    {
                        throw new FileNotFoundException("Failed to find file: " + filePath);
                    }
                }
                else
                {
                    var cached = this.CachedFiles[filePath];

                    cached.HitCount++;
                    cached.LastAccess = DateTime.Now;

                    return cached.Content;
                }
            }
        }

        /// <summary>
        /// Caches a new file with the content, if the file was already cached this will just count as an access and overwrite the content.
        /// </summary>
        /// <param name="filePath"></param>
        /// <param name="content"></param>
        public void Cache(string filePath, byte[] content)
        {
            lock (this.CachedFiles)
            {
                if (!this.CachedFiles.ContainsKey(filePath))
                {
                    if ((this.BytesCached + content.Length) > this.MaxBytesCached)
                        this.Free((this.BytesCached + content.Length) - this.MaxBytesCached);

                    var cached = new CachedFile(filePath, content);

                    this.BytesCached += content.Length;

                    this.CachedFiles.Add(filePath, cached);
                }
                else
                {
                    var cached = this.CachedFiles[filePath];

                    cached.HitCount++;
                    cached.LastAccess = DateTime.Now;
                    cached.Content = content;
                }
            }
        }

        /// <summary>
        /// Attempts to free up space in the cache by a specific amount of bytes.
        /// </summary>
        /// <param name="neededBytes"></param>
        public void Free(int neededBytes)
        {
            while (this.BytesCached + neededBytes >= this.MaxBytesCached)
            {
                CachedFile weakest = null;

                foreach (var pair in this.CachedFiles)
                {
                    if (weakest == null)
                        weakest = pair.Value;
                    else if (pair.Value.LastAccess < weakest.LastAccess && pair.Value.HitCount < weakest.HitCount)
                        weakest = pair.Value;
                }

                if (weakest != null)
                {
                    this.CachedFiles.Remove(weakest.FilePath);

                    this.BytesCached -= weakest.Content.Length;
                }
            }
        }
    }
}
